# flake8: noqa
# fmt: off

from ludwig.schema.metadata.parameter_metadata import ParameterMetadata

PREPROCESSING_METADATA = (
    {'computed_fill_value': ParameterMetadata(internal_only=True),
     }
)
